# -*- coding: utf-8 -*-
#
# ibptws/produto.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
from decimal import Decimal

import requests

from .config import conf
from .excecoes import ErroIdentificacao
from .excecoes import ErroProdutoNaoEncontrado


_Produto = namedtuple('_Produto',
        'codigo uf ex descricao nacional importado estadual')


class Produto(_Produto):
    """
    Resposta à consulta de produtos.

    .. warning::

        O atributo ``ex`` (exceção à regra aplicada no NCM) não é uma string,
        mas um número inteiro.

        Note também que os atributos ``nacional``, ``estadual`` e ``importado``
        são valores de ponto flutuante. Para obter objetos :class:`Decimal`,
        utilize os atributos :attr:`aliquota_nacional`,
        :attr:`aliquota_importado` e :attr:`aliquota_estadual`.

    """
        
    __slots__ = ()
    
    @property
    def aliquota_nacional(self):
        return Decimal(str(self.nacional))
        
    @property
    def aliquota_importado(self):
        return Decimal(str(self.importado))
    
    @property
    def aliquota_estadual(self):
        return Decimal(str(self.estadual))


def get_produto(codigo_ncm, excecao=0):
    """Consulta o servico de produtos, procurando pelo código NCM e,
    opcionalmente, pela exceção à regra aplicada no NCM.

    :param str codigo_ncm: Código NCM do produto a ser consultado.

    :param int excecao: **Opcional** Exceção à regra aplicada ao NCM.

    :return: Retorna uma instância de :class:`Produto` contendo os detalhes do
        NCM e dos valores aproximados dos tributos que incidem sobre ele.

    :rtype: ibptws.produto.Produto

    :raises ErroProdutoNaoEncontrato: se o NCM/exceção não forem encontrados.

    :raises ErroIdentificacao: se o token ou o CNPJ configurados tiverem
        expirado ou não estiverem corretos.
    """

    response = requests.get(conf.endpoint.produtos, params=dict(
            token=conf.token, cnpj=conf.cnpj, uf=conf.estado,
            codigo=codigo_ncm, ex=excecao))

    if response.status_code == requests.codes.ok:
        data = response.json()
        return Produto(**{k.lower():v for k,v in data.items()})

    elif response.status_code == requests.codes.not_found:
        raise ErroProdutoNaoEncontrado('NCM={!r}, EX={!r}'.format(
                codigo_ncm, excecao))

    elif response.status_code == requests.codes.forbidden:
        raise ErroIdentificacao('IBPT token={!r}, cnpj={!r}, UF={!r}'.format(
                conf.token, conf.cnpj, conf.estado))

    response.raise_for_status()



# -*- coding: utf-8 -*-
#
# ibptws/servico.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
from decimal import Decimal

import requests

from .config import conf
from .excecoes import ErroIdentificacao
from .excecoes import ErroServicoNaoEncontrado


_Servico = namedtuple('_Servico',
        'codigo uf descricao tipo nacional importado estadual municipal')

class Servico(_Servico):
    """
    Resposta à consulta de Serviços.

    .. warning::

        Note que os atributos ``nacional``, ``estadual``, ``municipal`` e
        ``importado`` são valores de ponto flutuante. Para obter objetos
        :class:`Decimal`, utilize os atributos :attr:`aliquota_nacional`,
        :attr:`aliquota_importado`, :attr:`aliquota_estadual` e
        :attr:`aliquota_municipal`.

    """
    
    __slots__ = ()
    
    @property
    def aliquota_nacional(self):
        return Decimal(str(self.nacional))
        
    @property
    def aliquota_importado(self):
        return Decimal(str(self.importado))
    
    @property
    def aliquota_estadual(self):
        return Decimal(str(self.estadual))
        
    @property
    def aliquota_municipal(self):
        return Decimal(str(self.municipal))
        

def get_servico(codigo_nbs):
    """Consulta pelos tributos do Serviço, procurando pelo código NBS/LC116.

    :param str codigo_nbs: Código NBS/LC116 do serviço a ser consultado.

    :return: Retorna uma instância de :class:`Servico` contendo os valores
        aproximados dos tributos que incidem sobre ele.

    :rtype: ibptws.servico.Servico

    :raises ErroServicoNaoEncontrato: se o NBS/LC116 não forem encontrados.

    :raises ErroIdentificacao: se o token ou o CNPJ configurados tiverem
        expirado ou não estiverem corretos.
    """

    response = requests.get(conf.endpoint.servicos, params=dict(
            token=conf.token, cnpj=conf.cnpj, uf=conf.estado,
            codigo=codigo_nbs))

    if response.status_code == requests.codes.ok:
        data = response.json()
        return Servico(**{k.lower():v for k,v in data.items()})

    elif response.status_code == requests.codes.not_found:
        raise ErroServicoNaoEncontrado('NBS/LC116={!r}'.format(codigo_nbs))

    elif response.status_code == requests.codes.forbidden:
        raise ErroIdentificacao('IBPT token={!r}, cnpj={!r}, UF={!r}'.format(
                conf.token, conf.cnpj, conf.estado))

    response.raise_for_status()



# -*- coding: utf-8 -*-
#
# ibptws/provisoes.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import redis

from .produtos import get_produto, Produto
from .servicos import get_servico, Servico


EXPIRA_EM_24H = 24 * 60 * 60


class ProvisaoBase(object):
    """
    Classe base para o provisionamento da consulta, tornando possível o
    armazenamento das consultas dos valores aproximados dos tributos em
    cache, acelerando a consulta para produtos e serviços recém consultados.
    
    Esta classe em particular não faz esse provisionamento, apenas fornece
    uma base para os métodos básicos que darão suporte ao provisionamento.
    Uma implementação de provisionamento mínima deverá sobrescrever os métodos
    :meth:`get_produto` e :meth:`get_servico`.
    
    .. versionadded:: 0.3
    """
    
    def get_produto(self, ncm, ncm_ex):
        """
        Obtém as alíquotas para um produto conforme seu NCM/SH e exceção.
        As implementações de provisionamento deverão sobrescrever este método
        para obter os dados do provisionamento ou do web services do IBPT,
        quando não houver um provisionamento disponível.
        
        :param str ncm: O código NCM/SH do produto.
        :param int ncm_ex: A exceção da NCM. Note que, diferentemente da
            função :func:`~ibptws.produtos.get_produto`, este argumento não é
            opcional. Informe ``0`` se não houver exceção.
            
        :return: Retorna os dados do produto solicitado.
        :rtype: ibptws.produtos.Produto
        """
        raise NotImplementedError()
        
    
    def get_servico(self, nbs):
        """
        Obtém as alíquotas para um serviço conforme seu código NBS.
        As implementações de provisionamento deverão sobrescrever este método
        para obter os dados do provisionamento ou do web services do IBPT,
        quando não houver um provisionamento disponível.
        
        :param str nbs: O código NBS do serviço.

        :return: Retorna os dados do serviço solicitado.
        :rtype: ibptws.servicos.Servico
        """
        raise NotImplementedError()
        
        
class SemProvisao(ProvisaoBase):
    """
    Implementa uma base não provisionada, para ser utilizada como padrão em
    serviços que poderiam ser provisionados se houvesse uma implementação de
    provisionamento adequada. É contraditório mas esta classe não faz nenhum
    provisionamento, apenas acessa o web services do IBPT quando forem
    solicitados produtos e serviços.
    
    .. versionadded:: 0.3
    """
    
    def get_produto(self, ncm, ncm_ex):
        return get_produto(ncm, ncm_ex)
        
    
    def get_servico(self, nbs):
        return get_servico(nbs)


class ProvisaoViaRedis(ProvisaoBase):
    """
    Implementa um provisionamento baseado em um servidor `Redis`_.
    Produtos e serviços solicitados serão obtidos do web services do IBPT e
    então provisionados, com uma expiração padrão de 24h de modo que, dentro
    desse prazo, toda solicitação feita para um produto ou serviço provisionados
    serão obtidos de um servidor Redis, ao invés de acessar o web services.
    
    .. versionadded:: 0.3
    """
    
    def __init__(self, redis=None, expires=EXPIRA_EM_24H, **kwargs):
        """
        Inicia uma instância de :class:`ProvisaoViaRedis`.
        
        :param redis: Uma instância do servidor Redis. Normalmente, deverá ser
            uma instância ``redis.StrictRedis``(veja `redis-py`_). Mas poderá
            ser qualquer implementação que se comporte da mesma maneira. Se
            este argumento não for informado, será criada uma instância de
            ``redis.StrictRedis``, cujos argumentos deverão ser os mesmos
            argumentos esperados pela classe ``StrictRedis``.
            
        :param int expires: Especifica o tempo que uma consulta à produto ou
            serviço deverá durar até que expire, exigindo que seja feita uma
            nova consulta ao web services. Em segundos. Padrão é 24 horas.

        """
        self._redis = redis
        self._expires = expires
        self._kwargs = kwargs
        
    
    def _connect(self):
        self._redis = redis.StrictRedis(**self._kwargs)
        
        
    def _sanear(self, classe_entidade, dados):
        entidade = classe_entidade(**{k.lower():v for k,v in dados.items()})
        return getattr(self, '_sanear_{}'.format(
                classe_entidade.__name__.lower()))(entidade)
                
                
    def _sanear_produto(self, produto):
        return produto._replace(
                codigo=produto.codigo.decode('utf-8'),
                uf=produto.uf.decode('utf-8'),
                descricao=produto.descricao.decode('utf-8'),
                ex=int(produto.ex),
                nacional=float(produto.nacional),
                importado=float(produto.importado),
                estadual=float(produto.estadual))
                
    
    def _sanear_servico(self, servico):
        return servico._replace(
                codigo=servico.codigo.decode('utf-8'),
                uf=servico.uf.decode('utf-8'),
                descricao=servico.descricao.decode('utf-8'),
                tipo=servico.tipo.decode('utf-8'),
                nacional=float(servico.nacional),
                importado=float(servico.importado),
                estadual=float(servico.estadual),
                municipal=float(servico.municipal))
                

    def _get(self, metodo, classe_entidade, chave, *args, **kwargs):
        if self._redis is None:
            self._connect()
        
        dados = self._redis.hgetall(chave)
        
        if dados:
            # compõe a entidade dos dados obtidos do provisionamento, porém,
            # dados provisionados no Redis são convertidos para strings, por
            # isso é necessário sanear os atributos da entidade resultante
            # convertendo para os tipos Python corretos...
            entidade = self._sanear(classe_entidade, dados)
        else:
            # não foi possível obter do provisionamento,
            # obtém do web services do IBPT...
            entidade = metodo(*args, **kwargs)
            # ...e provisiona os dados obtidos
            with self._redis.pipeline() as pipe:
                pipe.hmset(chave, unicode_to_str(entidade._asdict()))
                pipe.expire(chave, self._expires)
                pipe.execute()

        return entidade
        
        
    def get_produto(self, ncm, ncm_ex):
        chave = 'ncm:{}:{}'.format(ncm, ncm_ex)
        return self._get(get_produto, Produto, chave, ncm, ncm_ex)
        
    
    def get_servico(self, nbs):
        chave = 'nbs:{}'.format(nbs)
        return self._get(get_servico, Servico, chave, nbs)


def unicode_to_str(self, d):
    """
    Runs through dictionary keys, converting every unicode value to str.
    """
    convert = lambda v: v.encode('utf-8') if isinstance(v, unicode) else v
    return {k: convert(v) for k, v in d.items()}



# -*- coding: utf-8 -*-
#
# ibptws/calculadoras.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from decimal import Decimal

from .provisoes import SemProvisao


ZERO = Decimal('0')
CEM = Decimal('100')


class DeOlhoNoImposto(object):
    """
    Implementa uma calculadora para a Lei 12.741/2012. O propósito é acumular
    os valores de *n* consultas e fornecer o resultado do valor aproximado dos
    tributos em cada esfera aplicável. Por exemplo:
    
    .. sourcecode:: python
    
        >>> calculadora = DeOlhoNoImposto()     # doctest: +SKIP
        >>> for item in itens:                  # doctest: +SKIP
        ...     calculadora.produto(
        ...             item.produto.ncm,
        ...             item.produto.ncm_ex,
        ...             item.subtotal())
        ...
        >>> calculadora.carga_federal()         # doctest: +SKIP
        Decimal('6.73')
        >>> calculadora.carga.estadual()        # doctest: +SKIP
        Decimal('8.47')
        >>> calculadora.total_tributos()        # doctest: +SKIP
        Decimal('15.20')
        
    .. versionadded:: 0.3
        
    """
    
    def __init__(self, provisao=None):
        self._fed_nacionais = []
        self._fed_importados = []
        self._estaduais = []
        self._municipais = []
        self._total = ZERO
        self._provisao = provisao or SemProvisao()
        
    
    def reiniciar(self):
        """
        Reinicia a calculadora, zerando os produtos e serviços acumulados.
        """
        self._fed_nacionais[:] = []
        self._fed_importados[:] = []
        self._estaduais[:] = []
        self._municipais[:] = []
        self._total = ZERO
        
        
    def produto(self, ncm, ncm_ex, valor):
        """
        Acumula os valores aproximados dos tributos sobre o subtotal do produto
        identificado pelo código NCM/SH e exceção.
        
        :param str ncm: Código NCM/SH do produto.
        :param int ncm_ex: Exceção à regra NCM/SG.
        :param Decimal valor: Valor do subtotal do produto, a quantidade
            comercializada pelo valor unitário de venda.
        
        """
        p = self._provisao.get_produto(ncm, ncm_ex)
        self._fed_nacionais.append(valor * (p.aliquota_nacional / CEM))
        self._fed_importados.append(valor * (p.aliquota_importado / CEM))
        self._estaduais.append(valor * (p.aliquota_estadual / CEM))
        self._total += valor
        
    
    def servico(self, nbs, valor):
        """
        Acumula os valores aproximados dos tributos sobre o subtotal do serviço
        identificado pelo código NBS.
        
        :param str nbs: Código NBS do serviço.
        :param Decimal valor: Valor do subtotal do serviço.
        
        """
        s = self._provisao.get_servico(nbs)
        self._fed_nacionais.append(valor * (s.aliquota_nacional / CEM))
        self._fed_importados.append(valor * (s.aliquota_importado / CEM))
        self._estaduais.append(valor * (s.aliquota_estadual / CEM))
        self._municipais.append(valor * (s.aliquota_municipal / CEM))
        self._total += valor
        
        
    def carga_federal(self):
        """
        Retorna o valor aproximado dos tributos na esfera federal, entre os
        produtos nacionais e importados.
        
        :rtype: decimal.Decimal
        """
        return self.carga_federal_nacional() + self.carga_federal_importado()
        
    
    def carga_federal_nacional(self):
        """
        Retorna o valor aproximado dos tributos na esfera federal para produtos
        nacionais.
        
        :rtype: decimal.Decimal
        """
        return sum(self._fed_nacionais or [ZERO,])
        
    
    def carga_federal_importado(self):
        """
        Retorna o valor aproximado dos tributos na esfera federal para produtos
        importados.
        
        :rtype: decimal.Decimal
        """
        return sum(self._fed_importados or [ZERO,])
        
    
    def carga_estadual(self):
        """
        Retorna o valor aproximado dos tributos na esfera estadual.
        :rtype: decimal.Decimal
        """
        return sum(self._estaduais or [ZERO,])
        
    
    def carga_municipal(self):
        """
        Retorna o valor aproximado dos tributos na esfera municipal.
        :rtype: decimal.Decimal
        """
        return sum(self._municipais or [ZERO,])
        
    
    def total_tributos(self):
        """
        Retorna a soma dos valores aproximados dos tributos entre todos os
        produtos e servicos considerados.
        
        :rtype: decimal.Decimal
        """
        return sum([
                self.carga_federal(),
                self.carga_estadual(),
                self.carga_municipal(),])
                
    
    def total(self):
        """
        Retorna a soma dos subtotais dos produtos e servicos calculados.
        :rtype: decimal.Decimal
        """
        return self._total
        
    
    def percentual_sobre_total(self):
        """
        Retorna o percentual que o total dos tributos representa sobre o
        valor total dos produtos e serviços calculados.
        :rtype: decimal.Decimal
        """
        if self.total().is_zero():
            return ZERO
        return self.total_tributos() / self.total()



# -*- coding: utf-8 -*-
#
# ibptws/config.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class Endpoint(object):

    def __init__(self):

        self.produtos = 'http://iws.ibpt.org.br/api/Produtos'
        """Endereço para o web services de produtos."""

        self.servicos = 'http://iws.ibpt.org.br/api/Servicos'
        """Endereço para o web services de serviços."""


class Configuracoes(object):

    def __init__(self):

        self.endpoint = Endpoint()
        """Endereços para os web serviçes do IBPT."""

        self.token = ''
        """Token fornecido pelo IBPT para acesso ao web services."""

        self.cnpj = ''
        """CNPJ da empresa relacionada ao :attr:`token`."""

        self.estado = ''
        """Sigla do Estado (unidade federativa) do domicílio do :attr:`cnpj`."""


conf = Configuracoes()



# -*- coding: utf-8 -*-
#
# ibptws/excecoes.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class ErroIdentificacao(Exception):
    """Lançado quando não for possível identificar o token ou o CNPJ,
    resultando um erro HTTP 403 na consulta ao serviço.
    """


class ErroNaoEncontrado(Exception):
    """Erro de base para os erros que indicam HTTP 404 (not found)."""


class ErroProdutoNaoEncontrado(ErroNaoEncontrado):
    """Lançado quando a consulta do serviço de Produtos não for
    capaz de localizar o produto solicitado, resultando em um erro
    HTTP 404 na consulta ao serviço.
    """


class ErroServicoNaoEncontrado(ErroNaoEncontrado):
    """Lançado quando a consulta de Serviços não for capaz de localizar o
    serviço solicitado, resultando em um erro HTTP 404 na consulta.
    """
